(* $Id: uPervasives.ml,v 1.2 2004/04/06 22:12:40 yori Exp $ *)
(* Copyright 2002, 2003 Yamagata Yoriyuki. distributed with LGPL *)

type uchar = UChar.t
open OOChannel

module Conv = CharEncoding.Make (UTF8)
module UTF8NF = UNF.Make (UTF8)
module UTF8CaseMap = CaseMap.Make(UTF8)
module UTF8Col = UCol.Make (UTF8)

let io_normalization_mode = ref `Off
let newline_mode = ref `LF

let int_of_uchar u = UChar.uint_code u
let uchar_of_int n = UChar.chr_of_uint n
let validate_utf8 s = UTF8.validate s

let sprint_uchar u =
  let n = UChar.uint_code u in
  let n2 = n land 0xffff in
  let n1 = n lsr 16 in
  if n1 = 0 then 
    Printf.sprintf "\\u%04X" n2
  else
    Printf.sprintf "\\U%04X%04X" n1 n2

let escaped_uchar u =
  let n = int_of_uchar u in
  if n > 0x7f || n < 0 then
    sprint_uchar u
  else
    Char.escaped (Char.chr n)

let backslash = Char.code '\\'
let sq = Char.code '\''
let dq = Char.code '"'
let lf = Char.code '\n'
let cr = Char.code '\r'
let tab = Char.code '\t'
let backspace = Char.code '\b'

let bb =  "\\\\"
let bdq = "\\\""
let bsq =   "\\'"

let escaped_utf8 s =
  let buf = Buffer.create 0 in
  let proc u =
    let n = int_of_uchar u in
    if n > 0x7f || n < 0 then
      Buffer.add_string buf (sprint_uchar u)
    else
      Buffer.add_string buf (String.escaped (String.make 1 (Char.chr n)))
  in
  UTF8.iter proc s;
  Buffer.contents buf

let printer_utf8 f s =
  let b = UTF8.Buf.create 0 in
  UTF8.iter
    (fun u ->
      if UChar.uint_code u = 92  then
	UTF8.Buf.add_string b "\\\\"
      else if UChar.is_printable u then UTF8.Buf.add_char b u
      else
	let s = sprint_uchar u in
	UTF8.Buf.add_string b s)
    s;
  let s = UTF8.Buf.contents b in
  Format.fprintf f "\"%s\"" (Conv.encode CharEncoding.iso_c_locale s)

let printer_uchar f u =
  Format.fprintf f "'%s'"
    (if UChar.uint_code u = backslash then "\\\\"
    else if UChar.is_printable u then 
      let s = UTF8.init 1 (fun _ -> u) in
      Conv.encode CharEncoding.iso_c_locale s
    else
      sprint_uchar u)

class utf8_in_channel (oc : uchar obj_input_channel) =
  let buf = UTF8.Buf.create 0 in
  object (self)
    val mutable b = ""
    val mutable pos = 0
    method private fill =
      UTF8.Buf.clear buf;
      UTF8.Buf.add_char buf oc#get;
      b <- UTF8.Buf.contents buf
    method get =
      if String.length b > 0 then begin
	let u = UTF8.look b pos in
	pos <- UTF8.next b pos;
	u
      end else
	oc#get
    method close = oc#close
    method input_char =
      if String.length b = 0 then self#fill;
      let c = b.[pos] in
      pos <- pos + 1;
      if pos >= String.length b then begin b <- ""; pos <- 0 end;
      c
    method input s p len =
      if len <= 0 then () else
      (if String.length b = 0 then begin
	UTF8.Buf.clear buf;
	while Buffer.length buf < len do 
	  UTF8.Buf.add_char buf oc#get 
	done;
	b <- Buffer.contents buf end);
      let len = min (String.length b) len in
      String.blit b pos s p len;
      pos <- pos + len;
      if pos >= String.length b then begin b <- ""; pos <- 0 end;
      len
    method really_input s p len =
      let len' = self#input s p len in
      ignore (self#input s (p + len') (len - len'))
    method peek =
      if String.length b = 0 then self#fill;
      UTF8.look b pos
  end

class utf8_in_channel_of_in_channel
    ?normalization
    ?(encoding="iso-c-locale") 
    nc
    =
  let normalization : [ `Off | `NFC ] = 
    match normalization with
      None -> !io_normalization_mode
    | Some m -> m in
  let oc = 
    new CharEncoding.in_channel (CharEncoding.of_name encoding) nc in
  let oc = 
    match normalization with
      `Off -> oc
    | `NFC -> 
	new filter_in (new UNF.nfc) oc in
  utf8_in_channel oc

class utf8_out_channel (oc : uchar obj_output_channel) =
  let b = Buffer.create 0 in
  object (self)
    method put u =
      if Buffer.length b > 0 then self#flush;
      oc#put u
    method flush  =
      let s = Buffer.contents b in
      UTF8.iter oc#put s;
      Buffer.clear b;
      oc#flush
    method close =
      self#flush;
      oc#close;
    method output = Buffer.add_substring b
    method output_char = Buffer.add_char b
  end

class utf8_out_channel_of_out_channel
    ?normalization
    ?(encoding="iso-c-locale") 
    nc =
  let normalization : [ `Off | `NFC ] = 
    match normalization with
      None -> !io_normalization_mode
    | Some m -> m in
  let oc = new CharEncoding.out_channel (CharEncoding.of_name encoding) nc in
  let oc = 
    match normalization with
      `Off -> oc
    | `NFC -> new UNF.nfc oc in
  utf8_out_channel oc
