(* $Id: uTF8Unicode.ml,v 1.1 2004/04/06 22:12:39 yori Exp $ *)
(* Copyright 2002, 2003 Yamagata Yoriyuki. distributed with LGPL *)

open UPervasives

module UTF8NF = UNF.Make (UTF8)
module UTF8CaseMap = CaseMap.Make (UTF8)
module UTF8Col = UCol.Make (UTF8)

module NormalizationOff = struct

  module Pervasives = struct

    let open_out 
	?normalization
	?encoding
	filename =
      let nc = Pervasives.open_out filename in
      new utf8_out_channel_of_out_channel ?normalization ?encoding nc

    let open_out_gen
	open_flags mode
	?normalization
	?encoding
	filename =
      let nc = Pervasives.open_out_gen open_flags mode filename in
      new utf8_out_channel_of_out_channel ?normalization ?encoding nc

    let flush uc = uc#flush
    let output_char uc = uc#put
    let output_string uc s = UTF8.iter uc#put s
    let close_out uc = uc#close

    let open_in
	?normalization
	?encoding
	filename =
      let nc = Pervasives.open_in filename in
      new utf8_in_channel_of_in_channel ?normalization ?encoding nc

    let open_in_gen
	open_flags mode
	?normalization
	?encoding
	filename =
      let nc = Pervasives.open_in_gen open_flags mode filename in
      new utf8_in_channel_of_in_channel ?normalization ?encoding nc

    let input_char uc = uc#get

    let lf = UChar.of_char '\n'
    let cr = UChar.of_char '\r'
    let uls = UChar.chr_of_uint 0x2028
    let ups = UChar.chr_of_uint 0x2029
    let nel = UChar.chr_of_uint 0x0085

    let input_line uc =
      let b = UTF8.Buf.create 0 in
      let rec read () =
	let u = uc#get in
	if 
	  UChar.eq u lf || UChar.eq u nel || 
	  UChar.eq u uls || UChar.eq u ups 
	then ()
	else if UChar.eq u cr then
	  if UChar.eq uc#peek lf then ignore(uc#get) else ()
	else begin 
	  UTF8.Buf.add_char b u; read () 
	end in
      read ();
      UTF8.Buf.contents b

    let stdin = new utf8_in_channel_of_in_channel Pervasives.stdin
    let stdout = new utf8_out_channel_of_out_channel Pervasives.stdout
    let stderr= new utf8_out_channel_of_out_channel Pervasives.stderr

    let print_char u = stdout#put u
    let print_string s = UTF8.iter stdout#put s

    let output_newline ?newline uc =
      let newline = match newline with
	None -> !newline_mode | Some n -> n in
      match newline with
	`CR -> uc#put cr
      | `CRLF -> uc#put cr; uc#put lf
      | `LF -> uc#put lf
      | `ULS -> uc#put uls

    let print_newline ?newline () = output_newline ?newline stdout

    let print_endline ?newline s =
      print_string s;
      print_newline ()

    let prerr_char u = stderr#put u
    let prerr_string s = UTF8.iter stderr#put s
    let prerr_newline ?newline () = output_newline ?newline stderr
    let prerr_endline ?newline s =
      prerr_string s;
      prerr_newline ()

    let read_line () = input_line stdin
    let (^) = Pervasives.(^)
  end

  include Pervasives

  module Char = struct
    include UChar
    open UCharInfo

    let tolower1 = UCharInfo.load_to_lower1_tbl ()
    let toupper1 = UCharInfo.load_to_upper1_tbl ()
    let totitle1 = UCharInfo.load_to_title1_tbl ()
    let sp_casing = UCharInfo.load_conditional_casing_tbl ()

    let escaped = escaped_uchar

    let rec match_condition loc condition =
      match condition with
	`Not condition -> not (match_condition loc condition)
      | `Locale loc' -> Locale.contain loc' loc
      | _ -> false

    let match_conditions conditions loc =
      List.for_all (match_condition loc) conditions

    let lowercase ?(locale=Locale.current_locale Locale.LC_CTYPE) u =
      let rec search = function
	  {lower = [u]; title = _; upper = _; condition = conditions} 
	  :: _
	  when match_conditions conditions locale -> u
	| _ :: rest -> search rest
	| [] -> UCharTbl.get tolower1 u in
      search (UCharTbl.get sp_casing u)
	
    let uppercase ?(locale=Locale.current_locale Locale.LC_CTYPE) u =
      let rec search = function
	  {lower = _; title = _; upper = [u]; condition = conditions} 
	  :: _
	  when match_conditions conditions locale -> u
	| _ :: rest -> search rest
	| [] -> UCharTbl.get toupper1 u in
      search (UCharTbl.get sp_casing u)

    let titlecase ?(locale=Locale.current_locale Locale.LC_CTYPE) u =
      let rec search = function
	  {lower = _; title = [u]; upper = _; condition = conditions} 
	  :: _
	  when match_conditions conditions locale -> u
	| _ :: rest -> search rest
	| [] -> UCharTbl.get totitle1 u in
      search (UCharTbl.get sp_casing u)

    let unsafe_chr = unsafe_chr_of_uint
  end

  module String = struct 
    include String
    let get s i =
      let n = RawByte.Char.code s.[i] in
      if n >= 0x80 && n < 0xc2 then invalid_arg "UTF8Unicode.String.get" else
      UTF8.look s i
    let set = ()
    let init = UTF8.init
    let make len u = init len (fun _ -> u)
    let length = UTF8.length
    let next = UTF8.next
    let prev = UTF8.prev
    let move = UTF8.move
    let nth s n = UTF8.move s 0 n
    let bytelength = String.length
    let iter = UTF8.iter
    let escaped = escaped_utf8

    let rec index_from s p u =
      if UChar.eq u (UTF8.look s p) then p else
      let p' = UTF8.next s p in
      if p' >= String.length s then raise Not_found else
      index_from s p' u

    let index s u = index_from s 0 u

    let rec rindex_from s p u =
      if UChar.eq u (UTF8.look s p) then p else
      let p' = UTF8.prev s p in
      if p' < 0 then raise Not_found else
      rindex_from s p' u

    let rindex s u = rindex_from s (UTF8.prev s (String.length s)) u

    let contain_from s p u = 
      try ignore (index_from s p u); true with
	Not_found -> false

    let contain s u = contain_from s 0 u

    let rcontain_from s p u =
      try ignore (rindex_from s p u); true with
	Not_found -> false

    let nfd = UTF8NF.nfd
    let nfkd = UTF8NF.nfkd
    let nfc = UTF8NF.nfc
    let nfkc = UTF8NF.nfkc

    let uppercase ?(locale=Locale.current_locale Locale.LC_CTYPE) s =
      UTF8CaseMap.uppercase ~locale s

    let lowercase ?(locale=Locale.current_locale Locale.LC_CTYPE) s =
      UTF8CaseMap.lowercase ~locale s

    let titlecase ?(locale=Locale.current_locale Locale.LC_CTYPE) s =
      UTF8CaseMap.titlecase ~locale s

    let capitalize ?(locale=Locale.current_locale Locale.LC_CTYPE) s =
      if bytelength s <= 0 then s else
      let u = UTF8.look s 0 in
      let u = Char.titlecase ~locale u in
      let b = UTF8.Buf.create 0 in
      UTF8.Buf.add_char b u;
      Buffer.add_substring b s (UTF8.next s 0) (bytelength s - 1);
      Buffer.contents b

    let uncapitalize = lowercase

    let compare = UTF8.compare

    let compare_col
	?(locale=Locale.current_locale Locale.LC_COLLATE)
	?(strength=0)
	?(variable=`Shifted)
	(s1 : string) (s2 : string) =
      let prec =
	if strength = 1 then Some `Primary else
	if strength = 2 then Some `Secondary else
	if strength = 3 then Some `Tertiary else
	if strength = 4 then Some `Quaternary else None in
      UTF8Col.compare ~locale ?prec ~variable s1 s2

    let search_col
	?(locale=Locale.current_locale Locale.LC_COLLATE)
	?(strength=0)
	?variable
	(s1 : string) (s2 : string) i =
      let prec =
	if strength = 1 then Some `Primary else
	if strength = 2 then Some `Secondary else
	if strength = 3 then Some `Tertiary else
	if strength = 4 then Some `Quaternary else None in
      UTF8Col.search ~locale ?prec ?variable s1 s2 i
  end

  module Buffer = struct
    include Buffer
    let add_char = UTF8.Buf.add_char

    let rec add_channel b uc n =
      add_char b uc#get;
      add_channel b uc (n - 1)

    let output_buffer uc b =
      let s = UTF8.Buf.contents b in
      UTF8.iter uc#put s
  end
end

module NormalizationNFC = struct
  include NormalizationOff.Pervasives
  let input_line uc = UTF8NF.nfc (input_line uc)
  let read_line () = input_line stdin
  let (^) = UTF8NF.nfc_append

  module Char = NormalizationOff.Char

  module String = struct
    include NormalizationOff.String
    let concat pad sts =
      nfc (concat pad sts)
  end

  module Buffer = struct
    include NormalizationOff.Buffer
    let contents b = String.nfc (contents b)
  end
end

include NormalizationOff
